from sklearn.base import BaseEstimator, RegressorMixin


def _as_list_of_str(columns):
    """Return none, one or more columns as a list."""
    columns = columns if columns else []
    if isinstance(columns, str):
        columns = [columns]
    return columns


class CallableMixin:
    """Makes transformer callable."""

    def __call__(self, *args, **kwargs):
        return self.transform(*args, **kwargs)


class IdentityRegressor(BaseEstimator, RegressorMixin, CallableMixin):
    """Use to turn a Pipeline into an estimator."""

    def fit(self, X, y=None):
        return self

    def predict(self, X):
        return X

    def fit_predict(self, X, y=None):
        self.fit(X, y)
        return self.predict(X)



import warnings

from ..batches.single import row as single_row

MESSAGE = (
    "`timeserio.preprocessing.batches` module is deprecated, "
    "use `timeserio.batches` instead."
)
warnings.warn(MESSAGE, DeprecationWarning, stacklevel=2)

PandasBatchGenerator = single_row.RowBatchGenerator



import warnings

from ..pipeline.pipeline import *  # noqa

MESSAGE = (
    "`timeserio.preprocessing.base` module is deprecated, "
    "use `timeserio.pipeline` instead."
)
warnings.warn(MESSAGE, DeprecationWarning, stacklevel=2)



from .encoding import (FeatureIndexEncoder,  # noqa
                       StatelessOneHotEncoder,
                       StatelessPeriodicEncoder,
                       StatelessTemporalOneHotEncoder)
from .pandas import (PandasColumnSelector,  # noqa
                     PandasValueSelector,
                     PandasSequenceSplitter)
from .datetime import PandasDateTimeFeaturizer, LagFeaturizer  # noqa
from .aggregate import AggregateFeaturizer  # noqa



import os
import random
from typing import Iterable

import numpy as np
import tensorflow as tf
from keras import backend as K  # noqa
from keras.engine import Layer


def iterlayers(model: Layer) -> Iterable[Layer]:
    """
    Return iterable over all layers (and sub-layers) of a model.

    This works because a keras Model is a sub-class of Layer.
    Can be used for freezing/un-freezing layers, etc.
    """
    if hasattr(model, 'layers'):
        for layer in model.layers:
            yield from iterlayers(layer)
    else:
        yield model


def seed_random():
    """Seed all random number generators to ensure repeatable tests.

    Sets python, `numpy`, and `tensorflow` random seeds
    to a repeatable states. This is useful in tests, but should not be
    used in production.

    https://keras.io/getting-started/faq/#how-can-i-obtain-reproducible-results-using-keras-during-development
    """
    os.environ['PYTHONHASHSEED'] = '0'
    np.random.seed(42)
    random.seed(12345)
    session_conf = tf.ConfigProto(
        intra_op_parallelism_threads=1, inter_op_parallelism_threads=1
    )
    tf.set_random_seed(1234)
    sess = tf.Session(graph=tf.get_default_graph(), config=session_conf)
    K.set_session(sess)



"""Defaults and parameters used repeatedly in tests.

Attributes:
    n_jobs (int): default `n_jobs` used in scikit-learn pipelines.
        Set to 1 as parallelism is implemented
        over batches of data when needed.

"""

n_jobs = 1


class Columns:
    """Default column names used in tests.

    Attributes:
        datetime (str): name of the `datetime` column in tests
        id (str): name of the id (categorical) column in tests
        target (str): name of the (real-valued) regression target column

    """

    datetime = 'datetime'
    id = 'id'
    target = 'target'



import abc
import functools
from typing import Union

import numpy as np

from ... import ini
from ...preprocessing.pandas import array_to_dataframe
from ..utils import ceiling_division
from .base import BatchGenerator


class ForecastBatchGeneratorBase(BatchGenerator):
    """Generate batches of sequence forecast examples.

    Assume single continuous timeseries.
    """

    def __init__(
        self,
        *,
        df,
        batch_size: Union[None, int] = None,
        sequence_length=2,
        columns=None,
        sequence_columns=None,
        sequence_prefix='seq_',
        last_step_columns=None,
        last_step_prefix='end_of_',
        forecast_steps_min=1,
        forecast_steps_max=1,
    ):
        self.df = df
        self.batch_size = batch_size
        self.sequence_length = sequence_length
        self.columns = columns
        self.sequence_columns = sequence_columns
        self.sequence_prefix = sequence_prefix
        self.last_step_columns = last_step_columns
        self.last_step_prefix = last_step_prefix

        if sequence_columns:
            if not sequence_prefix:
                raise ValueError('`sequence_prefix` must be non-empty')

        if last_step_columns:
            if not (set(last_step_columns) <= set(sequence_columns)):
                raise ValueError('`last_step_columns` must be a subset of '
                                 '`sequence_columns`')
            if not last_step_prefix:
                raise ValueError('`last_step_prefix` must be non-empty')
            if last_step_prefix == sequence_prefix:
                raise ValueError('`last_step_prefix` must be '
                                 'different from `sequence_prefix`')

        self.forecast_steps_min = forecast_steps_min
        self.forecast_steps_max = forecast_steps_max

    @property
    def num_points(self):
        """Return number of rows in original timeseries."""
        return len(self.df)

    @abc.abstractclassmethod
    def __len__(self):
        raise NotImplementedError

    @abc.abstractclassmethod
    def batch_seq_start_indices(self, batch_idx):
        raise NotImplementedError

    def _get_sequence_values(self, column, start_indices):
        values = self.df[column].values
        cols = [
            values[start_indices + s] for s in range(self.sequence_length)
        ]
        seq_values = np.vstack(cols).T
        return seq_values

    def __getitem__(self, batch_idx):
        if not len(self):
            raise IndexError('Batch index out of range: Empty batch generator')
        batch_idx = batch_idx % len(self)
        start_indices = self.batch_seq_start_indices(batch_idx)
        batch_size = len(start_indices)
        end_indices = start_indices + self.sequence_length
        fc_indices = end_indices + np.random.randint(
            self.forecast_steps_min - 1, self.forecast_steps_max, batch_size
        )
        cols = self.columns if self.columns else []
        sequence_columns = self.sequence_columns or []
        last_step_columns = self.last_step_columns or []

        cols = cols + sequence_columns
        batch_df = self.df[cols].iloc[fc_indices].copy()
        batch_df.reset_index(drop=True, inplace=True)
        for column in sequence_columns:
            seq_values = self._get_sequence_values(
                column, start_indices
            )
            seq_col_name = self.sequence_prefix + column
            batch_df = array_to_dataframe(
                seq_values,
                column=seq_col_name,
                df=batch_df
            )
        for column in last_step_columns:
            seq_col_name = self.sequence_prefix + column
            last_step_col_name = self.last_step_prefix + column
            batch_df[last_step_col_name] = batch_df[seq_col_name].iloc[:, -1]

        return batch_df


class SamplingForecastBatchGenerator(ForecastBatchGeneratorBase):
    """Generate batches of sequence forecast examples.

    Assume single continuous timeseries.
    """

    def __init__(
        self,
        *,
        df,
        batch_size: Union[None, int] = None,
        sequence_length=2,
        columns=None,
        sequence_columns=None,
        sequence_prefix='seq_',
        last_step_columns=None,
        last_step_prefix='end_of_',
        forecast_steps_min=1,
        forecast_steps_max=1,
        oversampling=1,
    ):
        super().__init__(
            df=df,
            batch_size=batch_size,
            sequence_length=sequence_length,
            columns=columns,
            sequence_columns=sequence_columns,
            sequence_prefix=sequence_prefix,
            last_step_columns=last_step_columns,
            last_step_prefix=last_step_prefix,
            forecast_steps_min=forecast_steps_min,
            forecast_steps_max=forecast_steps_max,
        )
        self.oversampling = oversampling or 1

    @property
    def num_examples(self):
        """Return number of examples to yield in one epoch."""
        num_examples = (
            self.num_points -
            self.forecast_steps_max
        ) // self.sequence_length * self.oversampling
        return max(0, num_examples)

    @property
    def _eff_batch_size(self):
        return self.batch_size or self.num_examples

    def __len__(self):
        return ceiling_division(self.num_examples, self._eff_batch_size)

    def batch_seq_start_indices(self, batch_idx):
        start_indices = np.random.randint(
            low=0,
            high=(
                self.num_points + 1 -
                self.sequence_length -
                self.forecast_steps_max
            ),
            size=self._eff_batch_size
        )
        return start_indices


class SequenceForecastBatchGenerator(ForecastBatchGeneratorBase):
    """Generate batches of sequence forecast examples.

    Assume single continuous timeseries.
    """

    def __init__(
        self,
        *,
        df,
        batch_size: Union[None, int] = None,
        sequence_length=2,
        id_column=None,
        sequence_columns=[ini.Columns.datetime, ini.Columns.target],
        sequence_prefix='seq_',
        last_step_columns=[ini.Columns.datetime],
        last_step_prefix='end_of_',
        forecast_steps_min=1,
        forecast_steps_max=1,
        batch_offset=False,
        batch_offset_period=1,
        dt_column=ini.Columns.datetime,
        start_time=None,
    ):
        super().__init__(
            df=df,
            batch_size=batch_size,
            sequence_length=sequence_length,
            columns=[id_column] if id_column else None,
            sequence_columns=sequence_columns,
            sequence_prefix=sequence_prefix,
            last_step_columns=last_step_columns,
            last_step_prefix=last_step_prefix,
            forecast_steps_min=forecast_steps_min,
            forecast_steps_max=forecast_steps_max,
        )
        self.batch_offset = batch_offset
        self.batch_offset_period = batch_offset_period
        self.dt_column = dt_column
        self.start_time = start_time
        if self.sequence_length % self.batch_offset_period != 0:
            raise ValueError(f'sequence_length not divisible'
                             f' by batch_offset_period')

    @property  # type: ignore
    @functools.lru_cache(None)
    def first_index(self):
        if self.start_time is None:
            return 0
        times = self.df[self.dt_column].dt.time.values
        first_idx = np.argmax(times == self.start_time)
        if not first_idx and times[0] != self.start_time:
            raise ValueError(f'Start time {self.start_time} not found in df')
        return first_idx

    @property
    def num_examples(self):
        """Return number of examples to yield in one epoch."""
        num_examples = (
            self.num_points -
            self.forecast_steps_max -
            self.first_index
        ) // self.sequence_length
        if self.batch_offset:
            num_examples -= 1  # to deal with random offset
        return max(0, num_examples)

    @property
    def _eff_batch_size(self):
        return self.batch_size or self.num_examples

    def __len__(self):
        return ceiling_division(self.num_examples, self._eff_batch_size)

    def batch_seq_start_indices(self, batch_idx):
        start_indices = np.arange(
            self.sequence_length * self._eff_batch_size * batch_idx,
            self.sequence_length *
            min(self._eff_batch_size * (batch_idx + 1), self.num_examples),
            self.sequence_length
        )
        return self.first_index + start_indices + self.random_offset_value

    @property
    def random_offset_value(self):
        if not self.batch_offset:
            return 0
        offset = np.random.randint(
            low=0,
            high=ceiling_division(
                self.sequence_length,
                self.batch_offset_period
            )
        ) * self.batch_offset_period
        return offset

    def _get_sequence_values(self, column, start_indices):
        """Efficient implementation for consecutive sequences."""
        seq_values = self.df[column].values[
            start_indices[0]:start_indices[-1] + self.sequence_length
        ]
        seq_values = seq_values.reshape((-1, self.sequence_length))
        return seq_values



import logging
import time

import numpy as np
from keras.callbacks import Callback

logger = logging.getLogger(__name__)


def _now():
    """Return the current time, helper for mocking."""
    return time.time()


def _get_log_metrics(history, exclude=('epoch', 'batches')):
    metrics = [k for k in history.keys() if k not in exclude]
    return metrics


def _format_epoch_metric(
    history, metric='loss', fmt='.2e', errors=False, idx=-1, with_name=True
):
    if metric not in history:
        return KeyError(f"Metric '{metric}' not found in history.")
    mean_value = history[metric][idx]
    formatted_value = f"{mean_value:{fmt}}"
    if errors and 'batches' in history and metric in history['batches'][idx]:
        batch_values = history['batches'][idx][metric]
        batch_min, batch_max = np.min(batch_values), np.max(batch_values)
        batch_std = np.std(batch_values)
        formatted_errors = (
            f" +- {batch_std:{fmt}}"
            f" [{batch_min:{fmt}} .. {batch_max:{fmt}}]"
        )
    else:
        formatted_errors = ""
    if with_name:
        return f"{metric}: {formatted_value}{formatted_errors}"
    else:
        return f"{formatted_value}{formatted_errors}"


def _format_epoch_summary(history, fmt='.2e', idx=-1):
    metrics = _get_log_metrics(history)
    formatted_values = [
        _format_epoch_metric(history, metric=metric, fmt=fmt, idx=idx)
        for metric in metrics
    ]
    formatted_epoch = _format_epoch_metric(
        history, metric='epoch', fmt='d', idx=idx
    )
    formatted_values = [formatted_epoch] + formatted_values
    message = ' - '.join(formatted_values)
    return message


class HistoryLogger(Callback):
    """Log all history, including per-batch losses and metrics.

    Based on keras.callbacks.History
    """

    def __init__(self, batches=True):
        self.batches = batches
        super().__init__()

    def on_train_begin(self, logs=None):
        logger.info('Training started.')
        self.history = {}

    def on_epoch_begin(self, epoch, logs=None):
        batch_history = self.history.setdefault('batches', [])
        batch_history.append({})
        self.epoch_time_start = _now()

    def on_batch_end(self, batch, logs=None):
        if not self.batches:
            return
        batch_history = self.history['batches'][-1]
        for k, v in logs.items():
            batch_history.setdefault(k, []).append(v)

    def on_epoch_end(self, epoch, logs=None):
        logs = logs or {}
        elapsed = _now() - self.epoch_time_start
        logs.update(epoch=epoch)
        logs.update(epoch_duration=elapsed)
        for k, v in logs.items():
            self.history.setdefault(k, []).append(v)
        message = _format_epoch_summary(self.history)
        logger.info(message)

    def _log_training_metric_improvement(self, fmt='.2e'):
        metrics = _get_log_metrics(
            self.history, exclude=('batches', 'epoch', 'epoch_duration')
        )
        for metric in metrics:
            before = _format_epoch_metric(
                self.history, metric=metric, idx=0, errors=True
            )
            after = _format_epoch_metric(
                self.history,
                metric=metric,
                idx=-1,
                with_name=False,
                errors=True
            )
            logger.info(f"{metric}: {before} --> {after}")

    def _log_training_time(self, fmt='.2e'):
        times = self.history['epoch_duration']
        mean, std, total = np.mean(times), np.std(times), np.sum(times)
        min_duration, max_duration = np.min(times), np.max(times)
        logger.info(f"Total duration: {total:{fmt}} seconds")
        logger.info(
            f"Epoch duration: {mean:{fmt}} +- {std:{fmt}}"
            f" [{min_duration:{fmt}} .. {max_duration:{fmt}}] seconds"
        )

    def on_train_end(self, logs=None):
        num_epochs = len(self.history['epoch'])
        logger.info(f"Training finished in {num_epochs} epochs.")
        self._log_training_time()
        self._log_training_metric_improvement()


class TimeLogger(Callback):
    """
    Logs the times of training the network and per epoch.

    A summary of the total statistics is given when the network finishes
    the training process.
    """

    def __init__(self):
        super(Callback, self).__init__()
        self.start_training_time = 0
        self.epoch_times = []
        self.training_time = 0

    def on_train_begin(self, logs=None):
        self.start_training_time = _now()
        logger.info("TimeLogger on_train_begin")

    def on_train_end(self, logs=None):
        self.training_time = _now() - self.start_training_time
        logger.info("Timing: Training took %0.2f", self.training_time)
        mean, std, max_elapsed, min_elapsed = self._get_stats()
        logger.info(
            "Timing: Epochs summary mean/std/max/min: %0.5f/%0.5f/%0.5f/%0.5f",
            mean, std, max_elapsed, min_elapsed
        )

    def _get_stats(self):
        mean = np.mean(self.epoch_times)
        std = np.std(self.epoch_times)
        max_elapsed = np.max(self.epoch_times)
        min_elapsed = np.min(self.epoch_times)
        return mean, std, max_elapsed, min_elapsed

    def on_epoch_begin(self, epoch, logs=None):
        self.epoch_time_start = _now()

    def on_epoch_end(self, epoch, logs=None):
        elapsed = _now() - self.epoch_time_start
        self.epoch_times.append(elapsed)
        logger.info("Timing: Epoch %d took %0.5f", epoch, float(elapsed))



import numbers

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing.base import _transform_selected
from sklearn.utils.validation import check_is_fitted


class FeatureIndexEncoder(BaseEstimator, TransformerMixin):
    """Encode categorical features.

    Like sklearn.preprocessing.LabelEncoder, but for Features - works for 2D.
    See also sklearn.preprocessing.OrdinalEncoder.
    Return values between 0 and n_classes-1.
    """

    @property
    def n_classes_(self):
        check_is_fitted(self, 'classes_')
        return len(self.classes_)

    def fit(self, X, y=None):
        """Fit feature encoder."""
        self.classes_ = np.unique(X)
        return self

    def transform(self, X, y=None):
        """Transform labels to normalized encoding."""
        check_is_fitted(self, 'classes_')
        classes = np.unique(X)
        if len(np.intersect1d(classes, self.classes_)) < len(classes):
            diff = np.setdiff1d(classes, self.classes_)
            raise ValueError("y contains new labels: %s" % str(diff))
        return np.searchsorted(self.classes_, X)


class StatelessOneHotEncoder(OneHotEncoder):
    """Stateless version of sklearn.preprocessing.OneHotEncoder."""

    def __init__(
        self,
        *,
        n_features,
        n_values,
        dtype=np.float64,
        sparse=True,
        handle_unknown='error'
    ):
        if isinstance(n_values, str):
            raise ValueError('n_values must be specified explicitly!')
        super().__init__(
            categories=self.get_categories(n_features, n_values),
            dtype=dtype,
            sparse=sparse,
            handle_unknown=handle_unknown
        )
        self.n_features = n_features
        self.n_values = n_values
        self._init_stateless()

    @staticmethod
    def get_categories(n_features, n_values):
        if isinstance(n_values, int):
            n_values = [n_values] * n_features
        if len(n_values) != n_features:
            raise ValueError(
                "`n_values` be an int, or a list of length `n_features`."
            )
        categories = [
            range(_n_values)
            for _n_values in n_values
        ]
        return categories

    def _init_stateless(self):
        X = np.zeros((1, self.n_features))
        super().fit(X)

    @property
    def required_columns(self):
        return {None}

    def transformed_columns(self, input_columns):
        return {None}


class StatelessTemporalOneHotEncoder(StatelessOneHotEncoder):
    """Temporal version of StatelessOneHotEncoder.

    Groups columns by one-hot feature rather than temporal order.
    Requires constant `n_values` for all timesteps/features.
    """

    def __init__(
        self,
        *,
        n_features,
        n_values,
        dtype=np.float64,
        sparse=True,
        handle_unknown='error'
    ):
        if not isinstance(n_values, int):
            raise ValueError('n_values must be a single integer!')
        super().__init__(
            n_features=n_features,
            n_values=n_values,
            dtype=dtype,
            sparse=sparse,
            handle_unknown=handle_unknown
        )

    @staticmethod
    def _reshape_temporal(y, n_features):
        n_examples = y.shape[0]
        y = y.reshape((n_examples, n_features, -1))
        y = y.transpose(0, 2, 1)
        y = y.reshape((n_examples, -1))
        return y

    def transform(self, X):
        y = super().transform(X)
        return self._reshape_temporal(y, self.n_features)


class PeriodicEncoder(BaseEstimator, TransformerMixin):
    """Periodic feature encoder.

    Computes periodic features as
    `sin( 2 * np.pi * (X - phase) / period)`
    and
    `cos( 2 * np.pi * (X - phase) / period)`

    All `sin` features are stacked adjacently, followed by `cos` features
    and non-periodic features.

    Parameters:
        periodic_features : "all" or array of indices or mask
            Specify what features are treated as categorical.

            - 'all' (default): All features are treated as categorical.
            - array of indices: Array of categorical feature indices.
            - mask: Array of length n_features and with dtype=bool.

            Non-categorical features are always stacked
            to the right of the matrix.

        period: float or list of floats
            feature period, either scalar
            or list of values for each periodic feature

        phase: float or list of floats
            feature phase offset, as period

    """

    def __init__(
        self,
        *,
        periodic_features="all",
        period,
        phase=0,
    ):
        self.periodic_features = periodic_features
        self.period = period
        self.phase = phase

    def fit(self, X, y=None):
        self.n_features_ = X.shape[1]
        return self

    def _check_value_per_feature(self, value, n_features):
        """Check one value is available per feature."""
        if isinstance(value, numbers.Number):
            return value
        if len(value) != n_features:
            raise ValueError(
                f'Must provide one constant or a list of {n_features} values,'
                f' got list of {len(value)}'
            )
        return np.array(value)

    def _transform(self, X, y=None):
        """Assume periodic features only."""
        n_features = X.shape[1]
        period = self._check_value_per_feature(self.period, n_features)
        phase = self._check_value_per_feature(self.phase, n_features)
        X = 2 * np.pi * (X - phase) / period
        X_sin, X_cos = np.sin(X), np.cos(X)
        return np.hstack([X_sin, X_cos])

    def transform(self, X, y=None):
        check_is_fitted(self, 'n_features_')
        n_features = X.shape[1]
        if n_features != self.n_features_:
            raise ValueError(f'X has different shape than during fitting.'
                             f' Expected {self.n_features_} features,'
                             f' got {n_features}.')
        return _transform_selected(
            X,
            self._transform,
            dtype=X.dtype,
            selected=self.periodic_features,
            copy=True
        )

    @property
    def required_columns(self):
        return {None}

    def transformed_columns(self, input_columns):
        return {None}


class StatelessPeriodicEncoder(PeriodicEncoder):
    """Stateless version of PeriodicEncoder.

    Computes periodic features as
    `sin( 2 * np.pi * (X - phase) / period)`
    and
    `cos( 2 * np.pi * (X - phase) / period)`

    All `sin` features are stacked adjacently, followed by `cos` features
    and non-periodic features.

    Parameters:
        n_features: int
            number of columns in X

        periodic_features : "all" or array of indices or mask
            Specify what features are treated as categorical.

            - 'all' (default): All features are treated as categorical.
            - array of indices: Array of categorical feature indices.
            - mask: Array of length n_features and with dtype=bool.

            Non-categorical features are always stacked
            to the right of the matrix.

        period: float or list of floats
            feature period, either scalar
            or list of values for each periodic feature

        phase: float or list of floats
            feature phase offset, as period

    """

    def __init__(
        self,
        *,
        n_features,
        periodic_features="all",
        period,
        phase=0,
    ):
        super().__init__(
            periodic_features=periodic_features,
            period=period,
            phase=phase
        )
        self.n_features = n_features
        self._init_stateless()

    def _init_stateless(self):
        X = np.zeros((1, self.n_features))
        super().fit(X)

    @property
    def required_columns(self):
        return {None}

    def transformed_columns(self, input_columns):
        return {None}



import numpy as np
import pandas as pd
import sklearn.pipeline
from sklearn.base import BaseEstimator, TransformerMixin, clone
from sklearn.utils.validation import check_is_fitted

from .. import ini


class FeatureUnion(sklearn.pipeline.FeatureUnion):
    """Adds column checking for pandas transformers."""

    def __init__(
        self,
        transformer_list,
        n_jobs=ini.n_jobs,
        transformer_weights=None
    ):
        super().__init__(
            transformer_list=transformer_list,
            n_jobs=n_jobs,
            transformer_weights=transformer_weights
        )

    @property
    def required_columns(self):
        required_columns = set()
        for _, transformer in self.transformer_list:
            required_columns |= transformer.required_columns
        return required_columns

    def transformed_columns(self, input_columns):
        transformed_columns = set()
        for _, transformer in self.transformer_list:
            transformed_columns_new = transformer.transformed_columns(
                input_columns
            )
            if transformed_columns_new is not None:
                transformed_columns |= transformed_columns_new
        return transformed_columns or None


class Pipeline(sklearn.pipeline.Pipeline):
    """Adds column checking for pandas transformers."""

    def __init__(self, steps, memory=None):
        super().__init__(steps=steps, memory=memory)

    @property
    def required_columns(self):
        first_transformer = self.steps[0][1]
        if len(self.steps) == 1:
            if not hasattr(first_transformer, 'required_columns'):
                return {None}
            return first_transformer.required_columns
        remaining_pipeline = Pipeline(
            steps=self.steps[1:]
        )
        required_columns = first_transformer.required_columns
        for _ in [0, 1]:
            intermediate_columns = first_transformer.transformed_columns(
                list(required_columns)
            )
            missing_columns = (
                remaining_pipeline.required_columns - intermediate_columns
            )
            required_columns |= missing_columns
        if missing_columns:
            raise ValueError(f'Pipeline not valid: {missing_columns}')

        return required_columns

    def transformed_columns(self, input_columns):
        transformed_columns = input_columns
        for _, transformer in self.steps:
            if not hasattr(transformer, 'transformed_columns'):
                transformed_columns = {None}
            else:
                transformed_columns = \
                    transformer.transformed_columns(transformed_columns)
        return transformed_columns


class GroupedPipeline(BaseEstimator, TransformerMixin):
    """Apply a pipeline seperately to each section of a DataFrame.groupby.

    Parameters:
        groupby : string or array of strings
            The column name(s) to group the input dataframe by.

        pipeline : Pipeline or Pipeline-like object
            The pipeline to apply to each section. Just needs to implement fit
            and predict or transform.

        errors : "raise", "return_df" or "return_empty"
            Specify behaviour if a groupby key unseen is encountered at
            transform or predict time.
            "raise" will always raise an error, "return_empty" will raise if
            every group is an unseen key, and "return_df" will not raise.

            - "raise" (default): raise a KeyError
            - "return_df": Return a copy of the the input dataframe. Useful if
                the pipeline returns the original dataframe with extra columns
                - the extra columns will be null for the missing key group,
                provided at least one group was a fitted key.
            - "return_none": Return an empty dataframe / numpy array. Useful if
                the pipeline returns new columns only - the missing key group
                will be all null rows.

    """

    def __init__(self, groupby, pipeline, errors='raise'):
        self.groupby = groupby
        self.pipeline = pipeline
        self.errors = errors

    def fit(self, df, y=None):
        self.pipelines_ = {}
        self.pipelines_ = {key: self._fit_subdf(sub_df)
                           for key, sub_df in df.groupby(self.groupby)}
        return self

    def _fit_subdf(self, sub_df):
        return clone(self.pipeline).fit(sub_df)

    def transform(self, df):
        return self._call_pipeline(df, 'transform')

    def predict(self, df):
        return self._call_pipeline(df, 'predict')

    def fit_predict(self, df, y=None):
        return self.fit(df, y).predict(df)

    def _call_pipeline(self, df, fnc):
        check_is_fitted(self, 'pipelines_')
        self.one_transformed = False
        transformed = [self._call_pipeline_subdf(key, sub_df, fnc)
                       for key, sub_df in df.groupby(self.groupby)]
        if not self.one_transformed and self.errors == 'return_empty':
            raise KeyError('All keys missing in fitted pipelines')
        out = pd.concat(transformed).reindex(df.index)
        # Convert back to np.array if the pipeline returns a np.array
        if self.one_transformed and self.cast_to_numpy:
            return out.values
        return out

    def _call_pipeline_subdf(self, key, sub_df, fnc):
        try:
            out = getattr(self.pipelines_[key], fnc)(sub_df)
            # type(out) is the same for each subdf
            self.cast_to_numpy = type(out) is np.ndarray
            # If out is pd.DataFrame, its unchanged
            # If out is np.array, its a newly indexed pd.DataFrame
            out = pd.DataFrame(out)
            out.index = sub_df.index
            self.one_transformed = True
            return out
        except KeyError:
            if self.errors == 'return_df':
                return sub_df
            if self.errors == 'return_empty':
                return pd.DataFrame(index=sub_df.index)
            raise KeyError(f"Missing key {key} in fitted pipelines")

    @property
    def required_columns(self):
        groupby = [self.groupby] if type(self.groupby) is str else self.groupby
        return self.pipeline.required_columns | set(groupby)

    def transformed_columns(self, input_columns):
        return self.pipeline.transformed_columns(input_columns)



import numpy as np
import pandas as pd
import dask.dataframe as dd

from .. import ini

DEF_N = 48
DEF_EMB_DIM = 2
DEF_SEQ_LENGTH = 3
DEF_FREQ = '0.5H'


def mock_datetime_range(periods=DEF_N, start=None):
    """Sample datetime range with a half-hour period."""
    if not start:
        start = pd.datetime.now()
    return pd.date_range(start=start, freq=DEF_FREQ, periods=periods)


def _single_user_fit_df(
    periods=DEF_N,
    start_date=None,
    id=0,
    embedding_dim=DEF_EMB_DIM,
    seq_length=DEF_SEQ_LENGTH
):
    embeddings = {
        i: np.random.rand(periods)
        for i in range(embedding_dim)
    }
    dt_range = mock_datetime_range(periods, start=0)
    seq_dt = {
        i: dt_range + pd.Timedelta(i / 2, unit='h')
        for i in range(seq_length)
    }
    seq_usage = {
        i: np.random.rand(periods)
        for i in range(seq_length)
    }
    df = pd.concat({
        'embedding': pd.DataFrame(embeddings),
        f'seq_{ini.Columns.datetime}': pd.DataFrame(seq_dt),
        f'seq_{ini.Columns.target}': pd.DataFrame(seq_usage),
    }, axis=1)
    df[ini.Columns.datetime] = mock_datetime_range(periods, start=start_date)
    df['weather_temperature'] = 30 * np.random.rand(periods)
    df[ini.Columns.target] = np.random.rand(periods)
    df[ini.Columns.id] = id
    df['cluster'] = 13
    return df


def mock_fit_data(periods=DEF_N,
                  start_date=None,
                  ids=[0],
                  embedding_dim=DEF_EMB_DIM,
                  seq_length=DEF_SEQ_LENGTH):
    """Create example fit data in the tall DataFrame format."""
    user_dfs = [
        _single_user_fit_df(
            periods=periods,
            start_date=start_date,
            id=id,
            embedding_dim=embedding_dim,
            seq_length=seq_length
        )
        for id in ids
    ]
    df = pd.concat(user_dfs, axis=0)
    df.reset_index(inplace=True, drop=True)
    return df


def mock_dask_fit_data(
    periods=DEF_N,
    start_date=None,
    ids=[0],
    embedding_dim=DEF_EMB_DIM,
    seq_length=DEF_SEQ_LENGTH
):
    """Create example fit data as a dask DataFrame.

    DataFrame is partitioned by ID.
    """
    df = mock_fit_data(
        periods=periods,
        start_date=start_date,
        ids=ids,
        embedding_dim=embedding_dim,
        seq_length=seq_length
    )
    ddf = dd.from_pandas(df, chunksize=periods)
    return ddf


def _single_user_raw_df(
    periods=DEF_N,
    start_date=None,
    id=0,
):
    df = pd.DataFrame({
        ini.Columns.id: id,
        ini.Columns.datetime: mock_datetime_range(periods, start=start_date),
        ini.Columns.target: np.random.rand(periods)
    })
    return df


def mock_raw_data(periods=DEF_N,
                  start_date=None,
                  ids=[0]):
    """Create example raw data in the tall DataFrame format."""
    user_dfs = [
        _single_user_raw_df(
            periods=periods,
            start_date=start_date,
            id=id,
        )
        for id in ids
    ]
    df = pd.concat(user_dfs, axis=0)
    df.reset_index(inplace=True, drop=True)
    return df


def mock_dask_raw_data(
    periods=DEF_N,
    start_date=None,
    ids=[0]
):
    """Create example fit data as a dask DataFrame.

    DataFrame is partitioned by ID.
    """
    df = mock_raw_data(
        periods=periods,
        start_date=start_date,
        ids=ids,
    )
    ddf = dd.from_pandas(df, chunksize=periods)
    return ddf


def mock_predict_data(periods=DEF_N, start_date=None):
    """Create example predict data in the tall DataFrame format."""
    df = mock_fit_data(periods=periods, start_date=start_date)
    return df.drop('usage', axis=1)



import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from .. import ini
from .utils import _as_list_of_str, CallableMixin

from holidays import UnitedKingdom as HolidayCalender

DAYS_IN_YEAR = 365
HOURS_IN_DAY = 24
MINUTES_IN_HOUR = 60
SECONDS_IN_MINUTE = 60
MINUTES_IN_DAY = MINUTES_IN_HOUR * HOURS_IN_DAY

PEAK_INTERVAL = ('16:00', '19:00')
DAY_INTERVAL = ('06:00', '23:00')
MORNING_INTERVAL = ('5:00', '10:00')


def get_fractional_hour_from_series(series: pd.Series) -> pd.Series:
    """
    Return fractional hour in range 0-24, e.g. 12h30m --> 12.5.

    Accurate to 1 minute.
    """
    hour = series.dt.hour
    minute = series.dt.minute
    return hour + minute / MINUTES_IN_HOUR


def get_fractional_day_from_series(series: pd.Series) -> pd.Series:
    """
    Return fractional day in range 0-1, e.g. 12h30m --> 0.521.

    Accurate to 1 minute
    """
    fractional_hours = get_fractional_hour_from_series(series)
    return fractional_hours / HOURS_IN_DAY


def get_fractional_year_from_series(series: pd.Series) -> pd.Series:
    """
    Return fractional year in range 0-1.

    Accurate to 1 day
    """
    return (series.dt.dayofyear - 1) / 365


def get_is_holiday_from_series(series: pd.Series) -> pd.Series:
    """Return 1 if day is a UK public holiday.

    FixMe: may require region information (England/Wales/Scotland)
    FixMe: maybe move to geo-related features
    """
    years = series.dt.year.unique()
    return series.dt.date.isin(HolidayCalender(years=years)).astype(int)


def get_zero_indexed_month_from_series(series: pd.Series) -> pd.Series:
    """Return months, in the range 0-11."""
    return series.dt.month - 1


def get_is_weekday_from_series(series: pd.Series) -> pd.Series:
    """Return 1 if day is weekday."""
    dow = series.dt.dayofweek
    return dow.isin([0, 1, 2, 3, 4])


def get_time_is_in_interval_from_series(
        series: pd.Series,
        *,
        start_time,
        end_time,
) -> pd.Series:
    """Return if time of day is in a given time interval."""
    if isinstance(start_time, str):
        start_time = pd.to_datetime(start_time).time()
    if isinstance(end_time, str):
        end_time = pd.to_datetime(end_time).time()

    times = series.dt.time

    wrap = False
    if start_time > end_time:
        wrap = True
        start_time, end_time = end_time, start_time

    start_condition = times >= start_time
    end_condition = times < end_time
    interval_condition = start_condition & end_condition
    if wrap:
        interval_condition = ~interval_condition

    return interval_condition


def get_is_peak_hour_from_series(series: pd.Series) -> pd.Series:
    """Return if time in peak hour interval."""
    interval = PEAK_INTERVAL
    return get_time_is_in_interval_from_series(series,
                                               start_time=interval[0],
                                               end_time=interval[1])


def get_is_daytime_from_series(series: pd.Series) -> pd.Series:
    """Return if time in daytime interval."""
    interval = DAY_INTERVAL
    return get_time_is_in_interval_from_series(series,
                                               start_time=interval[0],
                                               end_time=interval[1])


def get_is_morning_peak_from_series(series: pd.Series) -> pd.Series:
    """Return if time in morning peak interval."""
    interval = MORNING_INTERVAL
    return get_time_is_in_interval_from_series(series,
                                               start_time=interval[0],
                                               end_time=interval[1])


def truncate_series(series: pd.Series, truncation_period: str) -> pd.Series:
    return series.dt.to_period(truncation_period).dt.to_timestamp()


SUPPORTED_DATETIME_ATTRS = [
    'time',
    'hour',
    'month',
    'dayofweek',
    'dayofyear',
    'weekday_name',
]

# Custom datetime featurizers such as daylight etc.
CUSTOM_ATTRIBUTES = {
    'fractionalday': get_fractional_day_from_series,
    'fractionalhour': get_fractional_hour_from_series,
    'fractionalyear': get_fractional_year_from_series,
    'month0': get_zero_indexed_month_from_series,
    'is_holiday': get_is_holiday_from_series,
    'is_weekday': get_is_weekday_from_series,
    'is_in_interval': get_time_is_in_interval_from_series,
    'is_peak': get_is_peak_hour_from_series,
    'is_daytime': get_is_daytime_from_series,
    'is_morningpeak': get_is_morning_peak_from_series,
    'dt_truncated': truncate_series
}


class PandasDateTimeFeaturizer(BaseEstimator, TransformerMixin, CallableMixin):
    """Featurize datetime column by adding specified attributes."""

    valid_attributes = (
        SUPPORTED_DATETIME_ATTRS + list(CUSTOM_ATTRIBUTES.keys())
    )

    def __init__(
        self,
        column=ini.Columns.datetime,
        attributes=['month0', 'dayofweek', 'fractionalday'],
        kwargs=None
    ):
        self.column = column
        self.attributes = attributes
        self.kwargs = kwargs

    @property
    def attributes_(self):
        """Check attributes to be featurized."""
        return _as_list_of_str(self.attributes)

    def fit(self, df, y=None):
        return self

    def transform(self, df):
        if not isinstance(df, pd.DataFrame):
            raise TypeError('Input must be a DataFrame.')
        df = df.copy()
        column = df[self.column]
        if isinstance(column, pd.DataFrame):  # hierarchical index
            column_df = df[[self.column]].stack()
            new_features_df = self.transform(column_df).unstack()
            new_features_df = new_features_df[self.attributes_]
            df = pd.concat([df, new_features_df], axis=1)
            return df
        for attr in self.attributes_:
            if attr in SUPPORTED_DATETIME_ATTRS:
                df[attr] = column.dt.__getattribute__(attr)
            elif attr in CUSTOM_ATTRIBUTES:
                df[attr] = CUSTOM_ATTRIBUTES[attr](
                    column, **self.kwargs if self.kwargs else {}
                )
            else:
                raise KeyError(
                    f'Unknown attribute "{attr}"; '
                    'see `PandasDateTimeFeaturizer.valid_attributes`'
                )
        return df

    @property
    def required_columns(self):
        return set([self.column])

    def transformed_columns(self, input_columns):
        input_columns = set(_as_list_of_str(input_columns))
        if not self.required_columns <= input_columns:
            raise ValueError(f'Required columns are {self.required_columns}')
        return input_columns | set(self.attributes_)


class LagFeaturizer(BaseEstimator, TransformerMixin, CallableMixin):
    """Add lag features to a DataFrame.

    Note that None will be inserted where no historic data is available.

    Parameters:
        datetime_column: str
        columns: str or List[str]
            Feature column to lag according to datetime_column.
            See pandas.DataFrame.groupby
        lags: List[DateOffset, tseries.offsets, timedelta, or str]
            List of lags to apply.
            See pandas.DataFrame.shift
        duplicate_agg: str, default 'raise'
            Aggregation functions to apply to values for duplicate datetimes.
            By default, an error is raised if duplicates
            See pandas.DataFrame.groupby().aggregate()

    This Transformer is stateful.

    """

    def __init__(
        self,
        *,
        datetime_column,
        columns,
        lags,
        duplicate_agg='raise'
    ):
        self.datetime_column = datetime_column
        self.columns = columns
        self.lags = lags
        self.duplicate_agg = duplicate_agg

    def fit(self, df, y=None, **fit_params):
        columns = _as_list_of_str(self.columns)
        self.df_ = df.set_index(self.datetime_column)[columns]
        if self.duplicate_agg == 'raise':
            if any(self.df_.index.duplicated()):
                raise ValueError(
                    "Input dataframe contains duplicate entries "
                    "with the same %s", self.datetime_column
                )
        else:
            self.df_ = self.df_.groupby(level=0).agg(self.duplicate_agg)
        return self

    def transform(self, df):
        check_is_fitted(self, 'df_')
        lags = _as_list_of_str(self.lags)
        df = df.copy().set_index(self.datetime_column)
        for lag in lags:
            lag_df = self.df_.shift(freq=lag).add_suffix(f"_{lag}")
            df = pd.merge(
                df, lag_df,
                how="left", left_index=True, right_index=True,
                suffixes=("", "")
            )
        return df.reset_index()

    @property
    def required_columns(self):
        columns = [self.datetime_column] + _as_list_of_str(self.columns)
        return set(columns)

    def transformed_columns(self, input_columns):
        input_columns = set(_as_list_of_str(input_columns))
        if not self.required_columns <= input_columns:
            raise ValueError(f'Required columns are {self.required_columns}')
        lags = _as_list_of_str(self.lags)
        columns = _as_list_of_str(self.columns)
        new_columns = [f"{col}_{lag}" for lag in lags for col in columns]
        return input_columns | set(new_columns)



"""
A set of tools for determining output latency (delay and stride).

For temporal NN models.
"""
import keras.layers as kl
import numpy as np


def _n(num_or_tuple):
    """Get value from number or one-elt tuple/list."""
    try:
        num = num_or_tuple[0]
    except TypeError:
        num = num_or_tuple
    return num


def total_latency(relative_latencies):
    """
    Compute absolute delay + stride for last layer in a stack.

    Arguments:
    ----------
    relative_latencies: list of tuples
        each tuple is (rel_delay, rel_stride)

    Returns:
    --------
    delay: float
    stride: int

    """
    return absolute_latencies(relative_latencies)[-1]


def absolute_latencies(relative_latencies):
    """
    Compute absolute delay + stride for a each in a stack of layers.

    Arguments:
    ----------
    relative_latencies: list of tuples
        each tuple is (rel_delay, rel_stride)

    Returns:
    --------
    absolute_latencies: list of tuples
        each tuple is (abs_stride, abs_stride)

    """
    abs_delay, abs_stride = 0, 1
    absolute_latencies = []
    for idx, (rel_delay, rel_stride) in enumerate(relative_latencies):
        abs_delay += abs_stride * rel_delay
        abs_stride *= rel_stride
        absolute_latencies.append((abs_delay, abs_stride))
    return absolute_latencies


def layer_latency(layer):
    """
    Get latency for a given keras layer.

    Arguments:
    ----------
    layer: Layer
        a Keras layer

    Returns:
    --------
    delay: float
    stride: int

    """
    if isinstance(layer, kl.Conv1D):
        # https://keras.io/layers/convolutional/#conv1d
        stride = _n(layer.strides)
        if layer.padding == 'valid':
            delay = (_n(layer.kernel_size) - 1) * _n(layer.dilation_rate) / 2
        elif layer.padding == 'same':
            delay = 0
        else:
            raise ValueError('Padding {} not supported'.format(layer.padding))
    elif isinstance(layer, (kl.MaxPool1D, kl.AvgPool1D)):
        # https://keras.io/layers/pooling/#maxpooling1d
        stride = _n(layer.strides) if layer.strides else layer.padding
        if layer.padding == 'valid':
            delay = (_n(layer.pool_size) - 1) / 2
        elif layer.padding == 'same':
            delay = 0
    else:
        delay, stride = 0, 1

    return delay, stride


def model_latency(model):
    """
    Get latency for a given keras temporal model.

    Arguments:
    ----------
    model: keras.Model
        a Keras multi-layer temporal model

    Returns:
    --------
    delay: float
    stride: int

    """
    relative_latencies = [layer_latency(layer) for layer in model.layers]
    delay, stride = total_latency(relative_latencies)
    return delay, stride


def model_output_length(model, input_shape) -> int:
    """
    Get number of output timesteps for keras temporal model.

    Arguments:
    ----------
    model: keras.Model
        a Keras multi-layer temporal model

    input_shape: tuple
        shape of a single input

    """
    try:
        len(input_shape)
    except TypeError:
        input_shape = (input_shape, )

    if len(input_shape) == 1:
        batch_shape = (1, input_shape[0], 1)
    elif len(input_shape) == 2:
        batch_shape = (1, input_shape[0], input_shape[1])
    elif len(input_shape) == 3:
        batch_shape = input_shape
    else:
        raise ValueError(
            'input_shape must specify shape of single sample or batch, '
            'but found too many dimensions'
        )

    input_ = np.zeros(batch_shape)
    output = model.predict(input_)

    output_length = output.shape[1]
    return output_length



import abc
import contextlib
import copy
from typing import Optional, Dict, Collection
import warnings

from keras.models import Sequential, Model
from keras.callbacks import Callback
from keras.utils.generic_utils import has_arg

from ..utils.functools import get_default_args
from .utils import iterlayers
from .callbacks import HistoryLogger


class MultiNetworkBase(abc.ABC):
    """Abstract base class for models that contain multiple keras sub-models.

    Examples:
        Define a simple pair of regression models using `MultiNetworkBase`:

        >>> from timeserio.keras.multinetwork import MultiNetworkBase
        >>> from keras.models import Model
        >>> from keras.layers import Input, Dense

        >>> class MyNetwork(MultiNetworkBase):
        ...     def _model(self, hidden_units=8):
        ...         input = Input((1,))
        ...         intermediate = Dense(hidden_units)(input)  # shared layer
        ...         output_1 = Dense(1)(intermediate)
        ...         output_2 = Dense(1)(intermediate)
        ...         model_1 = Model(input, output_1)
        ...         model_2 = Model(input, output_2)
        ...         model_1.compile(loss="MSE", optimizer="SGD")
        ...         model_2.compile(loss="MSE", optimizer="SGD")
        ...         return {"model_1": model_1, "model_2": model_2}

        Instantiate a multi-network, using keyword arguments to provide
        hyperparameters:

        >>> mnet = MyNetwork(hidden_units=32)

        Models are instantiated on demand:

        >>> mnet.model
        {'model_1': <keras.engine.training.Model ...>, 'model_2': <...>}


    """

    def __init__(self, **hyperparams):
        self._model_instance = None
        self.hyperparams = hyperparams
        self.weights = None
        self.optimizers_config = None
        self.history = []

    # Abstract methods that need sub-classing #
    @abc.abstractmethod
    def _model(self, **kwargs) -> Dict[str, Model]:
        """Build and return a list of keras models."""
        raise NotImplementedError

    def _callbacks(self, **kwargs) -> Dict[str, Callback]:
        """Build and return a list of keras callbacks."""
        return {}

    # Hyperparameters #
    # Helper functions ##
    @property
    def _funcs_with_legal_params(self):
        return [
            Sequential.fit, Sequential.fit_generator, Sequential.predict,
            Sequential.predict_classes, Sequential.evaluate, self._model,
            self._callbacks
        ]

    @property
    def _funcs_with_default_params(self):
        return [self._model, self._callbacks]

    @property
    def _default_params(self):
        """Get default network parameters."""
        params = {'verbose': 0}
        for fn in self._funcs_with_default_params:
            params.update(get_default_args(fn))
        return params

    def _filter_hyperparams(self, fn, override=None):
        """Filter `hyperparams` and return those in `fn`'s arguments.

        Args:
            fn : arbitrary function
            override: dictionary, values to override `hyperparams`

        Returns:
            res : dictionary containing variables
                in both `hyperparams` and `fn`'s arguments.

        """
        override = override or {}
        res = {}
        for name, value in self.hyperparams.items():
            if has_arg(fn, name):
                res.update({name: value})
        res.update(override)
        return res

    def check_params(self, params) -> None:
        """Check for user typos in `params`.

        Args:
            params: dictionary; the parameters to be checked

        Raises:
            ValueError: if any member of `params` is not a valid argument.

        """
        for params_name in params:
            for fn in self._funcs_with_legal_params:
                if has_arg(fn, params_name):
                    break
            else:
                raise ValueError(f'{params_name} is not a legal parameter')

    # Public hyperparams setters/getters ##
    @property
    def hyperparams(self):
        params = self._default_params
        hyperparams = copy.deepcopy(self._hyperparams)
        params.update(hyperparams)
        return params

    @hyperparams.setter
    def hyperparams(self, params):
        self.check_params(params)
        if not hasattr(self, '_hyperparams'):
            self._hyperparams = {}
        self._hyperparams.update(params)

    def get_params(self, **params):
        """Get parameters for this estimator.

        Args:
            **params: ignored (exists for API compatibility).

        Returns:
            Dictionary of parameter names mapped to their values.

        """
        return self.hyperparams

    def set_params(self, **params):
        """Set the parameters of this estimator.

        Args:
            **params: Dictionary of parameter names mapped to their values.

        Returns:
            self

        """
        self.hyperparams = params
        return self

    # Model #
    def _init_model(self, reset_weights=False, reset_optimizers=False):
        self._model_instance = self._model(
            **self._filter_hyperparams(self._model)
        )
        if not reset_weights and self.weights:
            self._set_weights(self.weights)
        self.weights = self._get_weights()

        if not reset_optimizers and self.optimizers_config:
            self._set_optimizers_config(self.optimizers_config)
        self.optimizers_config = self._get_optimizers_config()

        return self._model_instance

    @property
    def model(self):
        if not hasattr(self, '_model_instance') or not self._model_instance:
            self._init_model()
        return self._model_instance

    @property
    def model_names(self):
        return list(self.model.keys())

    def check_model_name(self, model):
        if model not in self.model_names:
            raise ValueError(
                f'Unknown sub-model name {model}: '
                f'must be one of {self.model_names}'
            )

    @property
    def callbacks(self):
        callbacks = {model: [] for model in self.model}
        callbacks.update(
            self._callbacks(**self._filter_hyperparams(self._callbacks))
        )
        return callbacks

    # Weights ##
    def _get_weights(self) -> Dict:
        """Get a dictionary of weights per keras model."""
        weights = {
            name: model.get_weights()
            for name, model in self.model.items()
        }
        return weights

    def _set_weights(self, weights: Dict):
        """Set model weights from dictionary of weights per model."""
        for name, model in self.model.items():
            model.set_weights(weights[name])

    # Optimizers ##
    def _get_optimizers_config(self) -> Dict:
        """
        Get optimizer parameters.
        """
        optimizers_config = {
            name: model.optimizer.get_config()
            if hasattr(model, 'optimizer') and model.optimizer else None
            for name, model in self.model.items()
        }
        return optimizers_config

    def _set_optimizers_config(self, optimizers_config: Dict):
        """
        Set optimizer parameters.
        """
        for name, config in optimizers_config.items():
            try:
                self.model[name].optimizer.__init__(**config)
            except (AttributeError, TypeError):
                pass

    # Model freezing and training #
    @property
    def trainable_models(self):
        try:
            self._trainable_models
        except AttributeError:
            self._trainable_models = self.model_names
        return self._trainable_models

    @trainable_models.setter
    def trainable_models(self, value):
        MESSAGE = (
            "`MultiNetworkBase.trainable_models` attribute is deprecated, "
            "use `trainable_models` keyword argument in "
            ".fit() or .fit_generator() instead."
        )
        warnings.warn(MESSAGE, DeprecationWarning, stacklevel=2)

        model_names = value or []
        if isinstance(model_names, str):
            model_names = [model_names]
        for model in model_names:
            self.check_model_name(model)
        self._trainable_models = value

    def _set_model_trainable(self, model=None, trainable=True):
        """Set trainable state for each layer in submodel."""
        self.check_model_name(model)
        submodel = self.model[model]
        for l in iterlayers(submodel):
            l.trainable = trainable

    def _compile_all_models(self):
        """Compile all submodels."""
        for model in self.model_names:
            submodel = self.model[model]
            try:
                submodel.compile(
                    loss=submodel.loss,
                    optimizer=submodel.optimizer,
                    metrics=submodel.metrics,
                )
            except AttributeError:
                pass

    def _freeze_models_except(self, model=None):
        """
        Freeze all layers of a given model except the layers given by 'model'.

        Args:
            model: string or list of strings

        """
        model = model or []
        if isinstance(model, str):
            model = [model]
        for m in self.model_names:
            self._set_model_trainable(model=m, trainable=False)
        for m in model:
            self._set_model_trainable(model=m, trainable=True)
        self._compile_all_models()

    def _freeze(self, trainable_models=None):
        """Freeze specified models.

        Use argument if a list is provided, else use self.trainable_models.
        """
        self._freeze_models_except(
            trainable_models if trainable_models is not None
            else self.trainable_models
        )

    def _unfreeze(self):
        """Set all keras models as trainable."""
        self._freeze_models_except(self.model_names)

    @contextlib.contextmanager
    def _model_context(
        self, *, reset_weights: bool, reset_optimizers: bool,
        reset_history: bool, freeze: bool, training: bool = False,
        persist_model: bool = False,
        trainable_models: Optional[Collection[str]] = None,
    ):
        try:
            if not persist_model or self._model_instance is None:
                self._init_model(
                    reset_weights=reset_weights,
                    reset_optimizers=reset_optimizers
                )
            if reset_history:
                self.history = []
            if freeze:
                self._freeze(trainable_models=trainable_models)
            yield
        finally:
            if freeze:
                self._unfreeze()
            if training:
                self.weights = self._get_weights()
                self.optimizers_config = self._get_optimizers_config()
            if not persist_model:
                self._model_instance = None

    def _training_context(
        self,
        *,
        reset_weights=False,
        reset_optimizers=True,
        reset_history=False,
        trainable_models=None,
    ):
        return self._model_context(
            reset_weights=reset_weights,
            reset_optimizers=reset_optimizers,
            reset_history=reset_history,
            freeze=True,
            training=True,
            persist_model=False,
            trainable_models=trainable_models,
        )

    def _prediction_context(self, persist_model=True):
        return self._model_context(
            reset_weights=False,
            reset_optimizers=False,
            reset_history=False,
            freeze=False,
            training=False,
            persist_model=persist_model
        )

    # History #
    def _add_history_record(self, *, model, history, trainable_models=None):
        """Add record of current training run to history."""
        record = {
            'model': model,
            'trainable_models': trainable_models or self.trainable_models,
            'history': history
        }
        self.history.append(record)

    # Fit/predict methods #
    def fit(
        self,
        x,
        y,
        model=None,
        reset_weights=False,
        reset_optimizers=True,
        reset_history=False,
        **kwargs
    ):
        """Run training on one model."""
        self.check_model_name(model)

        if 'trainable_models' in kwargs:
            trainable_models = kwargs.pop('trainable_models') or []
        else:
            trainable_models = None

        fit_args = self._filter_hyperparams(Sequential.fit)
        fit_args.update(callbacks=self.callbacks.get(model))
        fit_args.update(kwargs)

        history_cbk = HistoryLogger(batches=True)
        fit_args['callbacks'] += [history_cbk]

        training_context = self._training_context(
            reset_weights=reset_weights,
            reset_optimizers=reset_optimizers,
            reset_history=reset_history,
            trainable_models=trainable_models,
        )
        with training_context:
            self.model[model].fit(x, y, **fit_args)
        self._add_history_record(
            model=model, history=history_cbk.history,
            trainable_models=trainable_models
        )
        return history_cbk

    def fit_generator(
        self,
        generator,
        model=None,
        reset_weights=False,
        reset_optimizers=True,
        reset_history=False,
        **kwargs
    ):
        """Train model from `generator`.

        Args:
            generator : generator or keras.utils.Sequence
                Training batches as in Sequential.fit_generator
            init: bool
                initialize new model; use False to continue training
            **kwargs: dictionary arguments
                Legal arguments are the arguments of `Sequential.fit_generator`

        Returns:
            history : List[Dict]
                details about the training history at each epoch.

        """
        self.check_model_name(model)

        if 'trainable_models' in kwargs:
            trainable_models = kwargs.pop('trainable_models') or []
        else:
            trainable_models = None

        fit_args = self._filter_hyperparams(Sequential.fit_generator)
        fit_args.update(callbacks=self.callbacks.get(model))
        fit_args.update(kwargs)

        history_cbk = HistoryLogger(batches=True)
        fit_args['callbacks'] += [history_cbk]

        training_context = self._training_context(
            reset_weights=reset_weights,
            reset_optimizers=reset_optimizers,
            reset_history=reset_history,
            trainable_models=trainable_models,
        )
        with training_context:
            self.model[model].fit_generator(generator, **fit_args)
        self._add_history_record(
            model=model, history=history_cbk.history,
            trainable_models=trainable_models
        )
        return history_cbk

    def predict(self, x, model: str = None, **kwargs):
        """Return predictions for the given test data.

        Args:
            x: array-like, shape `(n_samples, n_features)`
                Test samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            **kwargs: dictionary arguments
                Legal arguments are the arguments of `Sequential.predict`.

        Returns:
            preds: array-like, shape `(n_samples, ...)`
                Predictions.

        """
        self.check_model_name(model)
        pred_kwargs = self._filter_hyperparams(Sequential.predict)
        pred_kwargs.update(kwargs)
        with self._prediction_context():
            predictions = self.model[model].predict(x, **pred_kwargs)
        return predictions

    def predict_generator(self, generator, model: str = None, **kwargs):
        """Return predictions from a batch generator.

        Args:
            generator : generator or keras.utils.Sequence
                predict batches as in Sequential.predict_generator
            **kwargs: dictionary arguments
                Legal arguments are the arguments of
                `Sequential.predict_generator`

        Returns:
            preds: array-like, shape `(n_samples, ...)`
                Predictions.

        """
        self.check_model_name(model)
        pred_kwargs = self._filter_hyperparams(
            Sequential.predict_generator, kwargs
        )
        with self._prediction_context():
            predictions = (
                self.model[model].predict_generator(generator, **pred_kwargs)
            )
        return predictions

    def evaluate(self, x, y, model=None, **kwargs):
        """
        Evaluate estimator on given data.

        Args:
            x:  array-like, shape `(n_samples, n_features)`
                Test samples where `n_samples` is the number of samples
                and `n_features` is the number of features.
            y: array-like
            model: string
            kwargs: keyword arguments.
                    Legal arguments are the arguments of `Sequential.evaluate`.

        Returns:
            metrics: List[float]
                A list of loss and evaluation metric values

        """
        self.check_model_name(model)
        ev_kwargs = self._filter_hyperparams(Sequential.evaluate, kwargs)
        with self._prediction_context():
            metrics = self.model[model].evaluate(x, y, **ev_kwargs)
        return metrics

    def score(self, *args, **kwargs):
        """Alias for scikit compatibility."""
        metrics = self.evaluate(*args, **kwargs)
        try:
            score = metrics[0]  # in case we have loss + multiple metrics
        except (TypeError, IndexError):
            score = metrics
        return score

    def evaluate_generator(self, generator, model=None, **kwargs):
        """
        Evaluate estimator on given data.

        Args:
            generator : generator or keras.utils.Sequence
                predict batches as in Sequential.evaluate_generator
            model: string
            **kwargs: keyword arguments.
                Legal arguments are the arguments of
                `Sequential.evaluate_generator`.

        Returns:
            evaluation metric value(s)

        """
        self.check_model_name(model)
        ev_kwargs = self._filter_hyperparams(Sequential.evaluate, kwargs)
        with self._prediction_context():
            evaluation = (
                self.model[model].evaluate_generator(generator, **ev_kwargs)
            )
        return evaluation

    # Pickling #
    def __getstate__(self) -> Dict:
        """Get a picklable state dictionary of the model."""
        state = {}
        state['hyperparams'] = self.hyperparams
        state['weights'] = self.weights
        state['optimizers'] = self.optimizers_config
        state['history'] = self.history
        return state

    def __setstate__(self, state: Dict):
        """Set model from unpickled state dictionary."""
        self.hyperparams = state.get('hyperparams')
        self.weights = state.get('weights')
        self.optimizers_config = state.get('optimizers')
        self.history = state.get('history')



from typing import Dict

from sklearn.base import BaseEstimator


class MultiPipeline(BaseEstimator):
    """Umbrella object for holding multiple pipelines.

    Pipelines are scikit-learn transformers.
    """

    def __init__(self, pipelines: Dict[str, BaseEstimator]) -> None:
        if not pipelines:
            pipelines = {}
        self.pipelines = pipelines

    def _get_param_names(self):
        """Get parameter names for the estimator."""
        return sorted(self.pipelines.keys())

    def __getattr__(self, item):
        if item in self.pipelines:
            return self.pipelines[item]
        else:
            # Default behaviour
            raise AttributeError

    def __getitem__(self, item):
        if item in self.pipelines:
            return self.pipelines[item]
        else:
            # Default behaviour
            raise KeyError(f'Unknown pipe "{item}"')

    @property
    def required_columns(self):
        required_columns = set()
        for pipeline in self.pipelines:
            required_columns |= self.pipelines[pipeline].required_columns
        return required_columns

    def transformed_columns(self, input_columns):
        transformed_columns = set()
        for pipeline in self.pipelines:
            transformed_columns |= self.pipelines[
                pipeline].transformed_columns(input_columns)
        return transformed_columns



import logging
from ..utils.pickle import loadf

logger = logging.getLogger()


PIPELINE_ERRORS = (ValueError, AttributeError)
PICKLE_ERRORS = (KeyError,)


def is_valid_transformer(transformer):
    """"""
    try:
        transformer.required_columns
        transformer.transformed_columns(transformer.required_columns)
        return True
    except PIPELINE_ERRORS:
        return False


def is_valid_pipeline(pipeline):
    """"""
    is_valid = is_valid_transformer(pipeline)
    returns_array = False
    if is_valid:
        transformed_columns = \
            pipeline.transformed_columns(pipeline.required_columns)
        if transformed_columns == {None}:
            returns_array = True
    return is_valid and returns_array


def is_valid_multipipeline(multipipeline):
    """"""
    all_valid = True
    for name, pipeline in multipipeline.pipelines.items():
        is_valid = is_valid_pipeline(pipeline)
        if not is_valid:
            logger.error('Invalid pipeline "%s"', name)
        all_valid &= is_valid
    return all_valid


def is_valid_multimodel(multimodel):
    """"""
    pipes_valid = is_valid_multipipeline(multimodel.multipipeline)
    return pipes_valid


def is_valid_pickle(path):
    """"""
    try:
        multimodel = loadf(path)
        return is_valid_multimodel(multimodel)
    except PICKLE_ERRORS:  # FixMe
        return False



import abc
from bisect import bisect_right

import numpy as np
import pandas as pd

from ..single.base import BatchGenerator
from ..utils import ceiling_division


class ChunkedBatchGenerator(BatchGenerator):
    """Base class for chunked/partitioned batch generators."""

    batch_aggregator = 1

    @abc.abstractmethod
    def __init__(self, batch_size=None, **kwargs):
        self.batch_size = batch_size
        self.__subgen_lengths = None

    @abc.abstractproperty
    def chunks(self):
        """Return an iterable of chunks.

        A chunk is passed to `self.make_subgen`
        """
        raise NotImplementedError

    @abc.abstractmethod
    def make_subgen(self, chunk) -> BatchGenerator:
        """Return a sub-generator for a single chunk.

        """
        raise NotImplementedError

    def _subgen_length(self, chunk):
        """Optimization for slow subgens."""
        if self.batch_size is None:
            return 1
        else:
            return len(self.make_subgen(chunk))

    @property
    def subgen_lengths(self):
        """List of lengths of individual sub-generators."""
        if self.__subgen_lengths is None:
            self.__subgen_lengths = [
                self._subgen_length(chunk)
                for chunk in self.chunks
            ]
        return self.__subgen_lengths

    @property
    def subgen_index_bounds(self):
        """Return List of subgen index bounds.

        First value is 0
        Last value is self.__len__()
        """
        return np.cumsum([0] + self.subgen_lengths)

    @property
    def num_subbatches(self):
        return sum(self.subgen_lengths)

    def __len__(self):
        return ceiling_division(self.num_subbatches, self.batch_aggregator)

    def find_subbatch_in_subgens(self, subbatch_idx):
        if not (0 <= subbatch_idx < self.num_subbatches):
            raise IndexError('Batch index out of range')
        subgen_idx = bisect_right(self.subgen_index_bounds, subbatch_idx) - 1
        idx_in_subgen = subbatch_idx - self.subgen_index_bounds[subgen_idx]
        return subgen_idx, idx_in_subgen

    def __getitem__(self, batch_idx):
        if not len(self):
            raise IndexError('Batch index out of range: Empty batch generator')
        batch_idx = batch_idx % len(self)
        subbatch_idx_start = batch_idx * self.batch_aggregator
        subbatch_idx_end = min(
            (batch_idx + 1) * self.batch_aggregator,
            self.num_subbatches
        )
        batch = pd.DataFrame()
        for subbatch_idx in range(subbatch_idx_start, subbatch_idx_end):
            subgen_idx, idx_in_subgen = self.find_subbatch_in_subgens(
                subbatch_idx
            )
            chunk = self.chunks[subgen_idx]
            subgen = self.make_subgen(chunk)
            subbatch = subgen[idx_in_subgen]
            batch = batch.append(subbatch, ignore_index=True)
        return batch



import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin

from .utils import _as_list_of_str


def array_to_dataframe(
    array: np.ndarray, column: str, df=None
) -> pd.DataFrame:
    """Add 1D or 2D array as column in df.

    If no df provided, return a new one.
    """
    if len(array.shape) < 0 or len(array.shape) > 2:
        raise ValueError('Expecting 1D or 2D array')
    if len(array.shape) == 1:
        array = array.reshape(-1, 1)
    n_columns = array.shape[1]
    cidx = pd.MultiIndex.from_arrays(
        [
            [column] * n_columns,
            range(n_columns),
        ]
    )
    df_arr = pd.DataFrame(array, columns=cidx)
    if df is None:
        df = df_arr
    else:
        df = _join_multilevel_dataframes([df, df_arr])
    return df


def _join_multilevel_dataframes(df_list):
    """Concat multiple dataframes.

    Support a combination of 1- and 2-deep indices.
    """
    minx_df = []
    for df in df_list:
        if isinstance(df.columns, pd.core.index.MultiIndex):
            minx_df.append(df)
        else:
            df.columns = pd.MultiIndex.from_product([df.columns, ['']])
            minx_df.append(df)
    # Join all dataframes together
    multi_concat = pd.concat(minx_df, axis=1)
    return multi_concat


class PandasColumnSelector(BaseEstimator, TransformerMixin):
    """Select a sub-set of columns from a pandas DataFrame."""

    def __init__(self, columns=None):
        self.columns = columns

    def fit(self, df, y=None, **fit_params):
        return self

    def transform(self, df):
        columns = _as_list_of_str(self.columns)
        subframe = df[columns]
        return subframe.copy()

    @property
    def required_columns(self):
        return set(_as_list_of_str(self.columns))

    def transformed_columns(self, input_columns):
        input_columns = set(_as_list_of_str(input_columns))
        if not self.required_columns <= input_columns:
            raise ValueError(f'Required columns are {self.required_columns}')
        return self.required_columns


class PandasValueSelector(BaseEstimator, TransformerMixin):
    """Select scalar - or vector-valued feature cols, and return np.array."""

    def __init__(self, columns=None):
        self.columns = columns

    def fit(self, df, y=None, **fit_params):
        return self

    def transform(self, df):
        columns = _as_list_of_str(self.columns)
        subarray = df[columns].values
        return subarray

    @property
    def required_columns(self):
        return set(_as_list_of_str(self.columns))

    def transformed_columns(self, input_columns):
        input_columns = set(_as_list_of_str(input_columns))
        if not self.required_columns <= input_columns:
            raise ValueError(f'Required columns are {self.required_columns}')
        return {None}


class PandasSequenceSplitter(BaseEstimator, TransformerMixin):
    """Split sequence columns in two."""

    def __init__(self, columns=None, index=0):
        self.columns = columns
        self.index = index

    def fit(self, df, y=None, **fit_params):
        return self

    def transform(self, df):
        columns = _as_list_of_str(self.columns)
        index = self.index
        for col in columns:
            values = df[col].values
            df = array_to_dataframe(values[:, :index], f'{col}_pre', df=df)
            df = array_to_dataframe(values[:, index:], f'{col}_post', df=df)
        return df

    @property
    def required_columns(self):
        columns = set(_as_list_of_str(self.columns))
        return columns

    def transformed_columns(self, input_columns):
        columns = _as_list_of_str(input_columns)
        input_columns = set(columns)
        if not self.required_columns <= input_columns:
            raise ValueError(f'Required columns are {self.required_columns}')
        to_change = _as_list_of_str(self.columns)
        columns = [f'{col}_post' for col in to_change]
        columns2 = [f'{col}_pre' for col in to_change]
        return set(np.concatenate([columns, columns2]))



from typing import Dict

from keras.utils import Sequence
from sklearn.base import BaseEstimator

from ..keras.multinetwork import MultiNetworkBase
from ..pipeline import MultiPipeline
from .pipegen import _PipelineGenerator


class MultiModel(BaseEstimator):
    """Multi-part model with pipelines."""

    def __init__(
        self, *, multinetwork: MultiNetworkBase, multipipeline: MultiPipeline,
        manifold: Dict
    ) -> None:
        self.multinetwork = multinetwork
        self.multipipeline = multipipeline
        self.manifold = manifold

    @property
    def model_names(self):
        return self.multinetwork.model_names

    @property
    def trainable_models(self):
        return self.multinetwork.trainable_models

    @trainable_models.setter
    def trainable_models(self, value):
        self.multinetwork.trainable_models = value

    @property
    def history(self):
        return self.multinetwork.history

    def get_model_pipes(self, model_name: str):
        x_pipe_names, y_pipe_names = self.manifold[model_name]
        if not x_pipe_names:
            x_pipe_names = []
        if not y_pipe_names:
            y_pipe_names = []
        if isinstance(x_pipe_names, str):
            x_pipe_names = [x_pipe_names]
        if isinstance(y_pipe_names, str):
            y_pipe_names = [y_pipe_names]
        x_pipes = [
            self.multipipeline[pipename] if pipename else None
            for pipename in x_pipe_names
        ]
        y_pipes = [
            self.multipipeline[pipename] if pipename else None
            for pipename in y_pipe_names
        ]
        return x_pipes, y_pipes

    @staticmethod
    def get_transformed(pipes, df):
        return [pipe.transform(df) for pipe in pipes]

    @staticmethod
    def get_fit_transformed(pipes, df):
        return [pipe.fit_transform(df) for pipe in pipes]

    def fit(self, df, model=None, **kwargs):
        """Fit one of the sub-networks from DataFrame."""
        x_pipes, y_pipes = self.get_model_pipes(model)
        x = self.get_fit_transformed(x_pipes, df)
        y = self.get_fit_transformed(y_pipes, df)
        if 'validation_data' in kwargs and \
                kwargs['validation_data'] is not None:
            val_df = kwargs['validation_data']
            val_x = self.get_transformed(x_pipes, val_df)
            val_y = self.get_transformed(y_pipes, val_df)
            kwargs['validation_data'] = (val_x, val_y)
        return self.multinetwork.fit(x, y, model=model, **kwargs)

    def predict(self, df, model=None, **kwargs):
        """Predict from one of the sub-networks."""
        x_pipes, _ = self.get_model_pipes(model)
        x = self.get_transformed(x_pipes, df)
        return self.multinetwork.predict(x, model=model, **kwargs)

    def predict_generator(self, df_generator, model=None, **kwargs):
        """Predict from one of the sub-networks on a DataFrame generator."""
        x_pipes, y_pipes = self.get_model_pipes(model)
        generator = _PipelineGenerator(
            x_pipes=x_pipes, y_pipes=y_pipes, df_generator=df_generator
        )
        return self.multinetwork.predict_generator(generator,
                                                   model=model,
                                                   **kwargs)

    def fit_generator(self, df_generator, model=None, **kwargs):
        """Fit one of the sub-networks from a DataFrame generator."""
        x_pipes, y_pipes = self.get_model_pipes(model)
        if 'validation_data' in kwargs:
            is_none = (kwargs['validation_data'] is None)
            is_sequence = isinstance(kwargs['validation_data'], Sequence)
            if is_none:
                pass
            elif is_sequence:  # validate from df generator
                val_df_gen = kwargs['validation_data']
                kwargs['validation_data'] = _PipelineGenerator(
                    x_pipes=x_pipes, y_pipes=y_pipes, df_generator=val_df_gen
                )
            else:  # validate from DataFrame
                val_df = kwargs['validation_data']
                val_x = self.get_fit_transformed(x_pipes, val_df)
                val_y = self.get_fit_transformed(y_pipes, val_df)
                kwargs['validation_data'] = (val_x, val_y)
        generator = _PipelineGenerator(
            x_pipes=x_pipes, y_pipes=y_pipes, df_generator=df_generator
        )
        return self.multinetwork.fit_generator(
            generator, model=model, **kwargs
        )

    def evaluate(self, df, model=None, **kwargs):
        """Evaluate a subnetwork from a DataFrame."""
        x_pipes, y_pipes = self.get_model_pipes(model)
        x = self.get_transformed(x_pipes, df)
        y = self.get_transformed(y_pipes, df)
        metrics = self.multinetwork.evaluate(x, y, model=model, **kwargs)
        return metrics

    def evaluate_generator(self, df_generator, model=None, **kwargs):
        """Evaluate a subnetwork from a DataFrame generator."""
        x_pipes, y_pipes = self.get_model_pipes(model)
        generator = _PipelineGenerator(
            x_pipes=x_pipes, y_pipes=y_pipes, df_generator=df_generator
        )
        metrics = self.multinetwork.evaluate_generator(
            generator,
            model=model,
            **kwargs
        )
        return metrics

    def freeze_models_except(self, model=None):
        """Freeze all subnetworks except the one(s) specified by 'model'.

        Raises DeprecationWarning
        """
        self.multinetwork.freeze_models_except(model)

    def freeze(self):
        """Freeze all models."""
        self.trainable_models = None

    def unfreeze(self):
        """Unfreeze all models."""
        self.trainable_models = self.model_names



from typing import Union

from .base import BatchGenerator
from ..utils import ceiling_division


class RowBatchGenerator(BatchGenerator):
    """Generate rows of pd.DataFrame in batches as sub-frames."""

    def __init__(
        self,
        *,
        df,
        batch_size: Union[None, int] = None,
        columns=None
    ):
        self.df = df
        self.batch_size = batch_size
        self.columns = columns

    @property
    def _eff_batch_size(self):
        return self.batch_size or len(self.df)

    def __len__(self):
        return ceiling_division(len(self.df), self._eff_batch_size)

    def __getitem__(self, batch_idx):
        if not len(self):
            raise IndexError('Batch index out of range: Empty batch generator')
        columns = self.columns
        if not columns:
            columns = self.df.columns
        batch_idx = batch_idx % len(self)
        batch_df = self.df[columns].iloc[
            batch_idx * self._eff_batch_size:
            batch_idx * self._eff_batch_size +
            self._eff_batch_size
        ]

        return batch_df



from keras.utils import Sequence


class _PipelineGenerator(Sequence):
    def __init__(self, *, x_pipes, y_pipes, df_generator):
        self.x_pipes = x_pipes
        self.y_pipes = y_pipes
        self.df_generator = df_generator

    def __len__(self):
        return len(self.df_generator)

    def __getitem__(self, item):
        df = self.df_generator[item]
        x = [pipe.transform(df) for pipe in self.x_pipes]
        y = [pipe.transform(df) for pipe in self.y_pipes]
        return x, y



import s3fs
import joblib.externals.cloudpickle as cp

__all__ = ['dumpf', 'loadf']


def open_url(filename, mode):
    """Open file from local drive or s3 bucket.

    S3 filename must start with `s3://`.
    """
    if filename.startswith('s3://'):
        s3 = s3fs.S3FileSystem()
        file = s3.open(filename, mode)
    else:
        file = open(filename, mode)
    return file


def dumpf(obj, filename):
    """Serialize object to file of given name."""
    with open_url(filename, 'wb') as file:
        cp.dump(obj, file)


def loadf(filename):
    """Load serialized object from file of given name."""
    with open_url(filename, 'rb') as file:
        obj = cp.load(file)
    return obj


def dumps(obj):
    """Serialize object to string."""
    return cp.dumps(obj)


def loads(s):
    """Load serialized object from string."""
    return cp.loads(s)



#!/usr/bin/env python
# -*- coding: utf-8 -*

import os

from setuptools import find_packages, setup

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))
# read __version__ attribute from version.py
with open("timeserio/version.py") as f:
    exec(f.read())
VERSION = __version__  # type: ignore  # noqa
# User README.md as long description
with open("README.md", encoding="utf-8") as f:
    README = f.read()

setup(
    name="timeserio",
    version=VERSION,
    description="Machine Learning and Forecasting tools",
    long_description=README,
    long_description_content_type="text/markdown",
    # Credentials
    author="Octopus Energy",
    author_email="nerds@octopus.energy",
    url="https://github.com/octoenergy/timeserio",
    license="MIT",
    # Package
    packages=find_packages(exclude=("tests",)),
    entry_points={"console_scripts": []},
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "joblib",
        "numpy",
        "pandas",
        "scikit-learn==0.20.3",
        "keras==2.2.4",
        "s3fs",
        "holidays",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
)



from keras.utils import Sequence


class ArrayBatchGenerator(Sequence):
    """Generate batches from X, y arrays.

    Mainly used for testing `fit_generator` and related functions
    """

    def __init__(self, x, y, batch_size=4):
        self.x = x
        self.y = y
        if len(self.y.shape) == 1:
            self.y = self.y.reshape((-1, 1))
        assert self.x.shape[0] == self.y.shape[0]
        self.batch_size = batch_size
        self.n_rows = self.x.shape[0]

    def __len__(self):
        return self.n_rows // self.batch_size + (
            1 if self.n_rows % self.batch_size else 0
        )

    def __getitem__(self, item):
        item = item % len(self)
        start_row = item * self.batch_size
        end_row = min((item + 1) * self.batch_size, self.n_rows)
        return self.x[start_row:end_row, :], self.y[start_row:end_row, :]



from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

from .utils import CallableMixin, _as_list_of_str


class AggregateFeaturizer(BaseEstimator, TransformerMixin, CallableMixin):
    """Create features as aggregates of all points sharing `groupby` keys.

    Parameters:
        groupby: str or List[str]
            Specify keys/column names to group by.
            See pandas.DataFrame.groupby

        aggregate: dict
            Aggregation functions to apply to each column after grouping.
            See pandas.DataFrame.groupby().aggregate()

    This Transformer is statefull, and the number of output columns depends
    on the input.

    """

    def __init__(self,
                 *,
                 aggregate: Dict[str, Union[str, List[str]]],
                 groupby: Optional[Union[str, List[str]]] = None):
        self.groupby = groupby
        self.aggregate = aggregate

    def fit(self, df, y=None):
        groupby = _as_list_of_str(self.groupby)
        to_aggregate = df.groupby(groupby) if groupby else df
        aggregated = self._agg(to_aggregate)
        aggregated.columns = _flatten_columns(aggregated.columns)
        self.df_ = aggregated
        return self

    def _agg(self, df):
        groupby = _as_list_of_str(self.groupby)
        aggregated = {}
        for column, methods in self.aggregate.items():
            methods = [methods] if type(methods) is str else methods
            for method in methods:
                single_agg = df[column].aggregate(method)
                if type(single_agg) is not pd.Series:
                    single_agg = [single_agg]
                aggregated[f'{column}_{method}'] = single_agg
        aggregated = pd.DataFrame(aggregated)
        if groupby:
            # Flatten the DF into a single row
            aggregated = pd.DataFrame(aggregated.unstack(groupby)).T
        return aggregated

    def transform(self, df):
        check_is_fitted(self, 'df_')
        return _cartesian_product(df, self.df_)

    @property
    def required_columns(self):
        required_columns = set(self.aggregate.keys())
        groupby = _as_list_of_str(self.groupby)
        required_columns |= set(groupby)
        return required_columns


def _flatten_columns(columns: pd.Index) -> Union[pd.Index, List[str]]:
    if type(columns) is not pd.MultiIndex:
        return columns
    height, width = np.array(columns.codes).shape
    new_cols = []
    for j in range(width):
        col = ''
        for i in range(height):
            name = columns.names[i]
            level = str(columns.levels[i][columns.codes[i][j]])
            col += f'{name}_{level}' if name else level
            col += '_'
        new_cols.append(col[:-1])
    return new_cols


def _cartesian_product(df_1: pd.DataFrame, df_2: pd.DataFrame) -> pd.DataFrame:
    df_1, df_2 = df_1.copy(), df_2.copy()
    df_1['dummy_key'] = 1
    df_2['dummy_key'] = 1
    return df_1.merge(df_2, on='dummy_key',
                      suffixes=('', '_agg')).drop('dummy_key', axis=1)



"""Generate batches from a data stored in Parquet files."""
import os

import s3fs
import pandas as pd

from .base import ChunkedBatchGenerator
from ... import ini
from ..single import row as single_row
from ..single import sequence as single_sequence


class RowBatchGenerator(ChunkedBatchGenerator):
    def __init__(
        self,
        *,
        path,
        batch_size=None,
        columns=None,
        batch_aggregator=1
    ):
        super().__init__()
        self.path = path
        self.batch_size = batch_size
        self.columns = columns
        self.batch_aggregator = batch_aggregator

        if self.path.startswith('s3://'):
            s3 = s3fs.S3FileSystem()
            self.files = [f's3://{file}' for file in s3.ls(self.path)]
        else:
            self.files = [
                os.path.join(path, file)
                for file in os.listdir(self.path)
            ]
        self.files = [f for f in self.files if f.endswith('.parquet')]

    @property
    def chunks(self):
        return self.files

    def make_subgen(self, chunk):
        filename = chunk
        subgen = single_row.RowBatchGenerator(
            df=pd.read_parquet(filename).reset_index(),
            batch_size=self.batch_size,
            columns=self.columns
        )
        return subgen


class SequenceForecastBatchGenerator(ChunkedBatchGenerator):
    def __init__(
        self,
        *,
        path,
        batch_size=None,
        sequence_length=2,
        id_column=ini.Columns.id,
        sequence_columns=[ini.Columns.datetime, ini.Columns.target],
        sequence_prefix='seq_',
        last_step_columns=[ini.Columns.datetime],
        last_step_prefix='end_of_',
        forecast_steps_min=1,
        forecast_steps_max=1,
        batch_offset=False,
        batch_offset_period=1,
        dt_column=ini.Columns.datetime,
        start_time=None,
        batch_aggregator=1
    ):
        super().__init__()
        self.path = path
        self.batch_size = batch_size
        self.sequence_length = sequence_length
        self.id_column = id_column
        self.sequence_columns = sequence_columns
        self.sequence_prefix = sequence_prefix
        self.last_step_columns = last_step_columns
        self.last_step_prefix = last_step_prefix
        self.forecast_steps_min = forecast_steps_min
        self.forecast_steps_max = forecast_steps_max
        self.batch_offset = batch_offset
        self.batch_offset_period = batch_offset_period
        self.dt_column = dt_column
        self.start_time = start_time
        self.batch_aggregator = batch_aggregator

        if self.path.startswith('s3://'):
            s3 = s3fs.S3FileSystem()
            self.files = [f's3://{file}' for file in s3.ls(self.path)]
        else:
            self.files = [
                os.path.join(path, file)
                for file in os.listdir(self.path)
            ]
        self.files = [f for f in self.files if f.endswith('.parquet')]

    @property
    def chunks(self):
        return self.files

    def make_subgen(self, chunk):
        filename = chunk
        subgen = single_sequence.SequenceForecastBatchGenerator(
            df=pd.read_parquet(filename).reset_index(),
            batch_size=self.batch_size,
            sequence_length=self.sequence_length,
            id_column=self.id_column,
            sequence_columns=self.sequence_columns,
            sequence_prefix=self.sequence_prefix,
            last_step_columns=self.last_step_columns,
            last_step_prefix=self.last_step_prefix,
            forecast_steps_min=self.forecast_steps_min,
            forecast_steps_max=self.forecast_steps_max,
            batch_offset=self.batch_offset,
            batch_offset_period=self.batch_offset_period,
            dt_column=self.dt_column,
            start_time=self.start_time,
        )
        return subgen
